@@ -5,11 +5,12 @@ from numbers import Number
 
 import numpy as np
 
+from joblib import Parallel, delayed
+
 from scipy.optimize import fmin_l_bfgs_b
 
 from sklearn.base import clone
 from sklearn.base import is_regressor
-from sklearn.externals.joblib import Parallel, delayed
 from sklearn.multioutput import MultiOutputRegressor
 from sklearn.utils import check_random_state
 
@@ -18,6 +19,7 @@ from ..acquisition import gaussian_acquisition_1D
 from ..learning import GaussianProcessRegressor
 from ..space import Categorical
 from ..space import Space
+from ..space.constraints import Constraints
 from ..utils import check_x_in_space
 from ..utils import cook_estimator
 from ..utils import create_result
@@ -68,6 +70,9 @@ class Optimizer(object):
         `x0` count as initialization points. If len(x0) < n_initial_points
         additional points are sampled at random.
 
+    * `lhs` [bool, default = False]:
+        If set to true the optimizer will use latin hypercube sampling for the first n_initial_points
+
     * `acq_func` [string, default=`"gp_hedge"`]:
         Function to minimize over the posterior distribution. Can be either
 
@@ -118,6 +123,11 @@ class Optimizer(object):
 
     * `acq_optimizer_kwargs` [dict]:
         Additional arguments to be passed to the acquistion optimizer.
+        options are:
+        - "length_scale_bounds" [list] a list of tuples with lower and upper bound
+        -  "length_scale" [list] a list of floats
+        - "n_restarts_optimizer" [int]
+        - "n_jobs" [int]
 
 
     Attributes
@@ -134,13 +144,13 @@ class Optimizer(object):
         to sample points, bounds, and type of parameters.
 
     """
+
     def __init__(self, dimensions, base_estimator="gp",
-                 n_random_starts=None, n_initial_points=10,
+                 n_random_starts=None, n_initial_points=10, lhs=False,
                  acq_func="gp_hedge",
                  acq_optimizer="auto",
                  random_state=None, acq_func_kwargs=None,
                  acq_optimizer_kwargs=None):
-
         self.rng = check_random_state(random_state)
 
         # Configure acquisition function
@@ -180,13 +190,28 @@ class Optimizer(object):
         self._n_initial_points = n_initial_points
         self.n_initial_points_ = n_initial_points
 
-        # Configure estimator
+        # record other arguments
+        if acq_optimizer_kwargs is None:
+            acq_optimizer_kwargs = dict()
 
+        self._length_scale_bounds = acq_optimizer_kwargs.get(
+            "length_scale_bounds", None)
+        self._length_scale = acq_optimizer_kwargs.get(
+            "length_scale", None)
+        self.n_points = acq_optimizer_kwargs.get("n_points", 10000)
+        self.n_restarts_optimizer = acq_optimizer_kwargs.get(
+            "n_restarts_optimizer", 5)
+        n_jobs = acq_optimizer_kwargs.get("n_jobs", 1)
+        self.n_jobs = n_jobs
+        self.acq_optimizer_kwargs = acq_optimizer_kwargs
+
+        # Configure estimator
+        self._check_length_scale_bounds(dimensions, self._length_scale_bounds)
         # build base_estimator if doesn't exist
         if isinstance(base_estimator, str):
             base_estimator = cook_estimator(
                 base_estimator, space=dimensions,
-                random_state=self.rng.randint(0, np.iinfo(np.int32).max))
+                random_state=self.rng.randint(0, np.iinfo(np.int32).max), length_scale_bounds=self._length_scale_bounds, length_scale=self._length_scale)
 
         # check if regressor
         if not is_regressor(base_estimator) and base_estimator is not None:
@@ -214,23 +239,12 @@ class Optimizer(object):
                              "'sampling', got {0}".format(acq_optimizer))
 
         if (not has_gradients(self.base_estimator_) and
-            acq_optimizer != "sampling"):
+                acq_optimizer != "sampling"):
             raise ValueError("The regressor {0} should run with "
                              "acq_optimizer"
                              "='sampling'.".format(type(base_estimator)))
         self.acq_optimizer = acq_optimizer
 
-        # record other arguments
-        if acq_optimizer_kwargs is None:
-            acq_optimizer_kwargs = dict()
-
-        self.n_points = acq_optimizer_kwargs.get("n_points", 10000)
-        self.n_restarts_optimizer = acq_optimizer_kwargs.get(
-            "n_restarts_optimizer", 5)
-        n_jobs = acq_optimizer_kwargs.get("n_jobs", 1)
-        self.n_jobs = n_jobs
-        self.acq_optimizer_kwargs = acq_optimizer_kwargs
-
         # Configure search space
 
         # normalize space if GP regressor
@@ -238,6 +252,14 @@ class Optimizer(object):
             dimensions = normalize_dimensions(dimensions)
         self.space = Space(dimensions)
 
+        # Latin hypercube sampling
+
+        self._lhs = lhs
+        if lhs:
+            self._lhs_samples = self.space.lhs(n_initial_points)
+
+        # Default is no constraints
+        self._constraints = None
         # record categorical and non-categorical indices
         self._cat_inds = []
         self._non_cat_inds = []
@@ -256,7 +278,7 @@ class Optimizer(object):
         # Initialize cache for `ask` method responses
 
         # This ensures that multiple calls to `ask` with n_points set
-        # return same sets of points. Reset to {} at every call to `tell`.
+        # return same sets of points. Reset to {} at every call to `tell` and `set_constraints`.
         self.cache_ = {}
 
     def copy(self, random_state=None):
@@ -272,6 +294,7 @@ class Optimizer(object):
             dimensions=self.space.dimensions,
             base_estimator=self.base_estimator_,
             n_initial_points=self.n_initial_points_,
+            lhs=self._lhs,
             acq_func=self.acq_func,
             acq_optimizer=self.acq_optimizer,
             acq_func_kwargs=self.acq_func_kwargs,
@@ -279,6 +302,11 @@ class Optimizer(object):
             random_state=random_state,
         )
 
+        # It is important to copy the constraints so that a call to '_tell()' will create a valid _next_x
+        optimizer._constraints = self._constraints
+        if self._lhs:
+            optimizer._lhs_samples = self._lhs_samples
+
         if hasattr(self, "gains_"):
             optimizer.gains_ = np.copy(self.gains_)
 
@@ -318,6 +346,7 @@ class Optimizer(object):
                flavours of `cl_x` strategies.
 
         """
+
         if n_points is None:
             return self._ask()
 
@@ -355,13 +384,16 @@ class Optimizer(object):
 
             if strategy == "cl_min":
                 y_lie = np.min(opt.yi) if opt.yi else 0.0  # CL-min lie
-                t_lie = np.min(ti) if ti is not None else log(sys.float_info.max)
+                t_lie = np.min(ti) if ti is not None else log(
+                    sys.float_info.max)
             elif strategy == "cl_mean":
                 y_lie = np.mean(opt.yi) if opt.yi else 0.0  # CL-mean lie
-                t_lie = np.mean(ti) if ti is not None else log(sys.float_info.max)
+                t_lie = np.mean(ti) if ti is not None else log(
+                    sys.float_info.max)
             else:
                 y_lie = np.max(opt.yi) if opt.yi else 0.0  # CL-max lie
-                t_lie = np.max(ti) if ti is not None else log(sys.float_info.max)
+                t_lie = np.max(ti) if ti is not None else log(
+                    sys.float_info.max)
 
             # Lie to the optimizer.
             if "ps" in self.acq_func:
@@ -385,8 +417,18 @@ class Optimizer(object):
         if self._n_initial_points > 0 or self.base_estimator_ is None:
             # this will not make a copy of `self.rng` and hence keep advancing
             # our random state.
-            return self.space.rvs(random_state=self.rng)[0]
 
+            assert not (
+                self._constraints and self._lhs), "Constraints can't be used while latin hypercube sampling is not exhausted"
+
+            if self._constraints:
+                # We use another sampling method when constraints are added
+                return self._constraints.rvs(random_state=self.rng)[0]
+            elif self._lhs:
+                # The samples are evaluated starting form lhs_samples[0]
+                return self._lhs_samples[len(self._lhs_samples)-self._n_initial_points]
+            else:
+                return self.space.rvs(random_state=self.rng)[0]
         else:
             if not self.models:
                 raise RuntimeError("Random evaluations exhausted and no "
@@ -477,7 +519,7 @@ class Optimizer(object):
         # after being "told" n_initial_points we switch from sampling
         # random points to using a surrogate model
         if (fit and self._n_initial_points <= 0 and
-           self.base_estimator_ is not None):
+                self.base_estimator_ is not None):
             transformed_bounds = np.array(self.space.transformed_bounds)
             est = clone(self.base_estimator_)
 
@@ -491,8 +533,13 @@ class Optimizer(object):
 
             # even with BFGS as optimizer we want to sample a large number
             # of points and then pick the best ones as starting points
-            X = self.space.transform(self.space.rvs(
-                n_samples=self.n_points, random_state=self.rng))
+            if self._constraints:
+                # We use another sampling method if constraints have been added
+                X = self.space.transform(self._constraints.rvs(
+                    n_samples=self.n_points, random_state=self.rng))
+            else:
+                X = self.space.transform(self.space.rvs(
+                    n_samples=self.n_points, random_state=self.rng))
 
             self.next_xs_ = []
             for cand_acq_func in self.cand_acq_funcs_:
@@ -501,8 +548,9 @@ class Optimizer(object):
                     acq_func=cand_acq_func,
                     acq_func_kwargs=self.acq_func_kwargs)
                 # Find the minimum of the acquisition function by randomly
-                # sampling points from the space
-                if self.acq_optimizer == "sampling":
+                # sampling points from the space. If constraints are present
+                # we use this strategy
+                if self.acq_optimizer == "sampling" or self._constraints:
                     next_x = X[np.argmin(values)]
 
                 # Use BFGS to find the mimimum of the acquisition function, the
@@ -586,3 +634,71 @@ class Optimizer(object):
 
         return create_result(self.Xi, self.yi, self.space, self.rng,
                              models=self.models)
+
+    def set_constraints(self, constraints):
+        ''' Sets the constraints for the optimizer
+
+        Parameters
+        ----------
+        * `constraints` [list] or [Constraints]:
+            Can either be a list of Constraint objects or a Constraints object
+        '''
+        if (self._n_initial_points >
+                0 and self._lhs):
+            raise RuntimeError(
+                "Can't set constraints while latin hypercube sampling points are not exhausted.")
+
+        if constraints:
+            if isinstance(constraints, Constraints):
+                # If constraints is a Constraints object we simply add it
+                self._constraints = constraints
+            else:
+                # If it is a list of constraints we initialize a Constraints object.
+                self._constraints = Constraints(constraints, self.space)
+        else:
+            self._constraints = None
+
+        self.update_next()
+
+    def remove_constraints(self):
+        ''' Sets constraints to None'''
+        self.set_constraints(None)
+
+    def get_constraints(self):
+        '''Returns constraints'''
+        return self._constraints
+
+    def update_next(self):
+        ''' Updates the value returned by opt.ask(). Useful if a parameter was updated after ask was called.'''
+        self.cache_ = {}
+        # Ask for a new next_x. Usefull if new constraints have been added or lenght_scale has been tweaked.
+        # We only need to overwrite _next_x if it exists.
+        if hasattr(self, '_next_x'):
+            opt = self.copy(random_state=self.rng)
+            self._next_x = opt._next_x
+
+    def get_result(self):
+        '''Returns the same result that would be returned by opt.tell() but without calling tell'''
+        return create_result(self.Xi, self.yi, self.space, self.rng,
+                             models=self.models)
+
+    def _check_length_scale_bounds(self, dimensions, bounds):
+        ''' Checks if length scale bounds are of in correct format'''
+        space = Space(dimensions)
+        n_dims = space.n_dims
+        if bounds:
+            if isinstance(bounds, list):
+                # Check that number of bounds is the same as number of dimensions
+                if not len(bounds) == n_dims:
+                    raise ValueError("Number of bounds (%s) must be same as number of dimensions (%s)" % (
+                        len(bounds), n_dims))
+                for i in range(len(bounds)):
+                    if not isinstance(bounds[i], tuple):
+                        raise TypeError("Each bound must be of type tuple, got %s" %
+                                        (type(bounds[i])))
+                    if not len(bounds[i]) == 2:
+                        raise ValueError("Each bound-tuple must have length 2, got %s" %
+                                         (len(bounds[i])))
+            else:
+                raise TypeError("Expected bounds to be of type list, got %s" %
+                                (type(bounds)))
