@@ -15,8 +15,9 @@ from .transformers import Identity
 from .transformers import Log10
 from .transformers import Pipeline
 
-
 # helper class to be able to print [1, ..., 4] instead of [1, '...', 4]
+
+
 class _Ellipsis:
     def __repr__(self):
         return '...'
@@ -92,7 +93,7 @@ def check_dimension(dimension, transform=None):
 
     if len(dimension) == 3:
         if (any([isinstance(dim, (float, int)) for dim in dimension[:2]]) and
-            dimension[2] in ["uniform", "log-uniform"]):
+                dimension[2] in ["uniform", "log-uniform"]):
             return Real(*dimension, transform=transform)
         else:
             return Categorical(dimension, transform=transform)
@@ -260,7 +261,7 @@ class Real(Dimension):
         return np.clip(
             super(Real, self).inverse_transform(Xt).astype(np.float),
             self.low, self.high
-            )
+        )
 
     @property
     def bounds(self):
@@ -295,6 +296,19 @@ class Real(Dimension):
                                "the space, not %s and %s." % (a, b))
         return abs(a - b)
 
+    def lhs_arange(self, n):
+        """ Returns an evenly distributed numpy array of samples to use with latin hypercube sampling.
+
+        Parameters
+        -----------
+        * `n` [int]
+            Number of samples.
+        """
+        a = (np.arange(n)+0.5)/n  # Evenly distributed betweeen 0 and 1
+
+        # Transform to the bounds of this dimension
+        return a*(self.high-self.low)+self.low
+
 
 class Integer(Dimension):
     def __init__(self, low, high, transform=None, name=None):
@@ -341,6 +355,9 @@ class Integer(Dimension):
             self._rvs = randint(self.low, self.high + 1)
             self.transformer = Identity()
 
+    def update_samplingspace(self, new_space):
+        self._rvs = new_space
+
     def __eq__(self, other):
         return (type(self) is type(other) and
                 np.allclose([self.low], [other.low]) and
@@ -387,6 +404,18 @@ class Integer(Dimension):
                                "the space, not %s and %s." % (a, b))
         return abs(a - b)
 
+    def lhs_arange(self, n):
+        """ Returns an evenly distributed list of samples to use with latin hypercube sampling.
+
+        Parameters
+        -----------
+        * `n` [int]
+            Number of samples.
+        """
+        rounded_numbers = np.round(np.linspace(self.low, self.high, n))
+        # convert to a list of integers
+        return [int(a) for a in rounded_numbers]
+
 
 class Categorical(Dimension):
     def __init__(self, categories, prior=None, transform=None, name=None):
@@ -410,7 +439,11 @@ class Categorical(Dimension):
         * `name` [str or None]:
             Name associated with dimension, e.g., "colors".
         """
-        self.categories = tuple(categories)
+        if transform == 'identity':
+            self.categories = tuple([str(c) for c in categories])
+        else:
+            self.categories = tuple(categories)
+
         self.name = name
 
         if transform is None:
@@ -423,7 +456,8 @@ class Categorical(Dimension):
             self.transformer = CategoricalEncoder()
             self.transformer.fit(self.categories)
         else:
-            self.transformer = Identity()
+            self.transformer = Identity(dtype=type(categories[0]))
+
         self.prior = prior
 
         if prior is None:
@@ -435,7 +469,7 @@ class Categorical(Dimension):
         # XXX check that sum(prior) == 1
         self._rvs = rv_discrete(
             values=(range(len(self.categories)), self.prior_)
-            )
+        )
 
     def __eq__(self, other):
         return (type(self) is type(other) and
@@ -457,7 +491,6 @@ class Categorical(Dimension):
 
     def rvs(self, n_samples=None, random_state=None):
         choices = self._rvs.rvs(size=n_samples, random_state=random_state)
-
         if isinstance(choices, numbers.Integral):
             return self.categories[choices]
         else:
@@ -505,6 +538,22 @@ class Categorical(Dimension):
                                " the space, not {} and {}.".format(a, b))
         return 1 if a != b else 0
 
+    def lhs_arange(self, n):
+        """ Returns an evenly distributed list of samples to use with latin hypercube sampling.
+
+        Parameters
+        -----------
+        * `n` [int]
+            Number of samples.
+        """
+
+        s = []
+        l = len(self.categories)  # Number of categories
+        for i in range(n):
+            # Loop through all categories by using the modulus.
+            s.append(self.categories[i % l])
+        return s
+
 
 class Space(object):
     """Search space."""
@@ -634,9 +683,9 @@ class Space(object):
         * `points`: [list of lists, shape=(n_points, n_dims)]
            Points sampled from the space.
         """
+
         rng = check_random_state(random_state)
 
-        # Draw
         columns = []
 
         for dim in self.dimensions:
@@ -647,7 +696,6 @@ class Space(object):
 
         # Transpose
         rows = []
-
         for i in range(n_samples):
             r = []
             for j in range(self.n_dims):
@@ -781,6 +829,11 @@ class Space(object):
         """Space contains exclusively categorical dimensions"""
         return all([isinstance(dim, Categorical) for dim in self.dimensions])
 
+    @property
+    def is_partly_categorical(self):
+        """Space contains any categorical dimensions"""
+        return any([isinstance(dim, Categorical) for dim in self.dimensions])
+
     def distance(self, point_a, point_b):
         """Compute distance between two points in this space.
 
@@ -797,3 +850,26 @@ class Space(object):
             distance += dim.distance(a, b)
 
         return distance
+
+    def lhs(self, n):
+        """ Returns n latin hypercube samples as a list of lists
+        """
+
+        samples = []
+        for i in range(self.n_dims):
+            lhs_perm = []
+            # Get evenly distributed samples form one dimension
+            lhs_aranged = self.dimensions[i].lhs_arange(n)
+            perm = np.random.permutation(n)
+            for p in perm:  # Random permutate the order of the samples
+                lhs_perm.append(lhs_aranged[p])
+            samples.append(lhs_perm)
+        # Now we have a list of lists with samples for each dimension.
+        # We need to transpose this so thatwe get a list of lists with samples for all the dimensions
+        transposed_samples = []
+        for i in range(n):
+            row = []
+            for j in range(self.n_dims):
+                row.append(samples[j][i])
+            transposed_samples.append(row)
+        return transposed_samples
